from typing import Callable, Hashable

from stupidb.comparable import Comparable
from stupidb.row import Row

PartitionBy = Callable[[Row], Hashable]
OrderBy = Callable[[Row], Comparable]
Preceding = Callable[[Row], int]
Following = Preceding

Predicate = Callable[[Row], bool]

Projector = Callable[[Row], Row]
JoinProjector = Callable[[Row, Row], Row]



import functools
from numbers import Real
from typing import (
    Callable,
    Generic,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    TypeVar,
)
from typing import Union as Union_

import toolz

import ibis.expr.datatypes as dt
import ibis.expr.schema as sch
from stupidb.stupidb import (
    AbstractAggregateSpecification,
    AggregateSpecification,
    CrossJoin,
    Difference,
    GroupBy,
    GroupingKeySpecification,
    InnerJoin,
    Intersection,
    JoinPredicate,
    Mean,
    OutputType,
    PopulationCovariance,
    Projection,
    Relation,
    Row,
    SampleCovariance,
    Selection,
    Sum,
    UnaryRelation,
    Union,
)
from stupidb.typehints import JoinProjector, Predicate, Projector


class RightShiftablePartial(functools.partial, Generic[OutputType]):
    def __rshift__(self, other: "RightShiftablePartial") -> Relation:
        return other(self)

    def __rrshift__(self, other: Relation) -> Relation:
        return self(other)

    def __iter__(self) -> Iterator[OutputType]:
        # XXX: Assumes all arguments have been bound
        # TODO: This seems a bit hacky. Refactor shifting.
        return iter(self())

    @property
    def columns(self) -> List[str]:
        return self().columns

    @property
    def schema(self) -> sch.Schema:
        return self().schema


V = TypeVar("V")


def table(
    rows: Iterable[Mapping[str, V]], schema: Optional[sch.Schema] = None
) -> RightShiftablePartial:
    first, rows = toolz.peek(rows)
    return RightShiftablePartial(
        UnaryRelation,
        child=((Row.from_mapping(row, id=id),) for id, row in enumerate(rows)),
        schema=(
            sch.Schema.from_dict(toolz.valmap(dt.infer, first))
            if schema is None
            else schema
        ),
    )


def cross_join(right: UnaryRelation) -> RightShiftablePartial:
    return RightShiftablePartial(CrossJoin, right=right)


def inner_join(
    right: UnaryRelation, predicate: JoinPredicate
) -> RightShiftablePartial:
    return RightShiftablePartial(InnerJoin, right=right, predicate=predicate)


def select(
    **projectors: Union_[
        Projector, JoinProjector, AbstractAggregateSpecification
    ]
) -> RightShiftablePartial:
    return RightShiftablePartial(Projection, projectors=projectors)


def sift(predicate: Predicate) -> RightShiftablePartial:
    return RightShiftablePartial(Selection, predicate=predicate)


def exists(relation: Relation) -> bool:
    return any(row for (row,) in relation)


def group_by(
    group_by: GroupingKeySpecification, aggregates: AggregateSpecification
) -> RightShiftablePartial:
    return RightShiftablePartial(
        GroupBy, group_by=group_by, aggregates=aggregates
    )


def union(right: Relation) -> RightShiftablePartial:
    return RightShiftablePartial(Union, right=right)


def intersection(right: Relation) -> RightShiftablePartial:
    return RightShiftablePartial(Intersection, right=right)


def difference(right: Relation) -> RightShiftablePartial:
    return RightShiftablePartial(Difference, right=right)


# Pull out the first element: all operations should ultimately call this


def do() -> RightShiftablePartial:
    return RightShiftablePartial(functools.partial(map, toolz.first))


# Aggregations


def sum(getter: Callable[[Row], Real]) -> AggregateSpecification:
    return AggregateSpecification(Sum, getter)


def mean(getter: Callable[[Row], Real]) -> AggregateSpecification:
    return AggregateSpecification(Mean, getter)


def samp_cov(
    arg1: Callable[[Row], Real], arg2: Callable[[Row], Real]
) -> AggregateSpecification:
    return AggregateSpecification(SampleCovariance, arg1, arg2)


def pop_cov(
    arg1: Callable[[Row], Real], arg2: Callable[[Row], Real]
) -> AggregateSpecification:
    return AggregateSpecification(PopulationCovariance, arg1, arg2)



import ast
import inspect
import typing
from typing import Iterator, Set, Tuple

import toolz

from .stupidb import JoinPredicate, Row, Rows


class AttributeFinder(ast.NodeVisitor):
    def __init__(self) -> None:
        self.keys: Set[str] = set()

    def visit_Subscript(self, node: ast.Subscript) -> None:
        index: ast.Index = typing.cast(ast.Index, node.slice)
        value: ast.Str = typing.cast(ast.Str, index.value)
        self.keys.add(value.s)


def find_attributes(predicate: JoinPredicate) -> Set[str]:
    finder = AttributeFinder()
    finder.visit(ast.parse(inspect.getsource(predicate)))
    return finder.keys


def merge(
    left: Rows, right: Rows, predicate: JoinPredicate
) -> Iterator[Tuple[Row, Row]]:
    attributes = find_attributes(predicate)
    sorter = toolz.flip(toolz.pluck)(attributes)
    left_sorted = sorted(left, key=sorter)
    right_sorted = sorted(right, key=sorter)
    left_iter = iter(left_sorted)
    right_iter = iter(right_sorted)
    left_row = next(left_iter)
    right_row = next(right_iter)
    while True:
        if predicate(left_row, right_row):
            yield left_row, right_row
            left_row = next(left_iter)
            right_row = next(right_iter)
        elif sorter(left) < sorter(right):
            left_row = next(left_iter)
        else:  # if left.Key > right.Key
            right_row = next(right_iter)



# -*- coding: utf-8 -*-

"""The stupidest database.

This is project designed to illustate the concepts that underly a typical
relational database implementation, starting at naive execution of table-stakes
features up to rule-based query optimization.

Please do not use this for any other reason than learning. There are no
guarantees here except that there will be bugs.

Plan of Attack
--------------
* Projection
* Selection
* Group By
* Joins

Requirements
------------
* The database must be able to operate on datasets that do not fit into main
  memory.
* The database must use as little memory as possible (within the bounds of
  standard Python code) and use only O(1) memory where possible.
* The database must produce rows using generators.
* The database must accept any iterable of mapping as its input(s).
* The database must attempt to use only built-in Python data structures
  including those in standard library modules.

"""

import abc
import collections
import functools
import itertools
import operator
import typing
from numbers import Real
from operator import methodcaller
from typing import (
    Any,
    Callable,
    Dict,
    FrozenSet,
    Generic,
    Hashable,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
)
from typing import Union as Union_

from typing_extensions import DefaultDict, NoReturn

import ibis.expr.schema as sch
from stupidb.row import Row
from stupidb.typehints import (
    Following,
    OrderBy,
    PartitionBy,
    Preceding,
    Predicate,
)

try:
    import cytoolz as toolz
except ImportError:
    import toolz as toolz


Rows = Iterable[Row]  # Rows are an Iterable of Row
InputType = TypeVar("InputType", Tuple[Row], Tuple[Row, Row])
OutputType = TypeVar("OutputType", Tuple[Row], Tuple[Row, Row])


class Relation(Generic[InputType, OutputType], metaclass=abc.ABCMeta):
    """A relation."""

    def __init__(self, child: Iterable[InputType], schema: sch.Schema) -> None:
        self.child: Iterable[InputType] = child
        self.schema = schema

    @property
    def columns(self) -> List[str]:
        return list(self.schema.names)

    @abc.abstractmethod
    def operate(self, args: InputType) -> OutputType:
        ...

    def __iter__(self) -> Iterator[OutputType]:
        for id, row in enumerate(filter(all, map(self.operate, self.child))):
            yield tuple(Row.from_mapping(element, id=id) for element in row)


class UnaryRelation(Relation[Tuple[Row], Tuple[Row]]):
    def operate(self, row: Tuple[Row]) -> Tuple[Row]:
        return row


class Projection(Relation[InputType, Tuple[Row]]):
    def __init__(
        self, child: Relation, projectors: Mapping[str, Callable[..., Row]]
    ) -> None:
        self.child = child
        self.projectors = projectors

    def operate(self, args: InputType) -> Tuple[Row]:
        mapping = {
            name: projector(*args)
            for name, projector in self.projectors.items()
        }
        row = Row(mapping)
        return (row,)


class Selection(UnaryRelation):
    def __init__(self, child: UnaryRelation, predicate: Predicate) -> None:
        self.child = child
        self.predicate = predicate

    def operate(self, row: Tuple[Row]) -> Tuple[Row]:
        result = self.predicate(*row)
        return row if result else (Row({}, id=row[0].id),)


GroupingKeySpecification = Mapping[str, Callable[[Row], Hashable]]
GroupingKey = Tuple[str, Hashable]
GroupingKeys = Tuple[GroupingKey, ...]

Input1 = TypeVar("Input1")
Input2 = TypeVar("Input2")
Output = TypeVar("Output")


class UnaryAggregate(Generic[Input1, Output], metaclass=abc.ABCMeta):
    def step(self, input1: Optional[Input1]) -> None:
        ...

    def finalize(self) -> Optional[Output]:
        ...


class UnaryWindowAggregate(UnaryAggregate[Input1, Output]):
    def inverse(self, input1: Optional[Input1]) -> None:
        ...

    def value(self) -> Optional[Output]:
        ...


class BinaryAggregate(Generic[Input1, Input2, Output]):
    def step(self, input1: Optional[Input1], input2: Optional[Input2]) -> None:
        ...

    def finalize(self) -> Optional[Output]:
        ...


Aggregate = Union_[UnaryAggregate, BinaryAggregate]


class FrameClause:
    def __init__(
        self,
        order_by: Sequence[OrderBy],
        partition_by: Sequence[PartitionBy],
        preceding: Optional[Preceding],
        following: Optional[Following],
    ) -> None:
        self._order_by = order_by
        self._partition_by = partition_by
        self._preceding = preceding
        self._following = following
        ...


class RowsMode(FrameClause):
    pass


class RangeMode(FrameClause):
    pass


class Window:
    @classmethod
    def rows(
        self,
        order_by: Sequence[OrderBy] = (),
        partition_by: Sequence[PartitionBy] = (),
        preceding: Optional[Preceding] = None,
        following: Optional[Following] = None,
    ) -> FrameClause:
        return RowsMode(order_by, partition_by, preceding, following)

    @classmethod
    def range(
        self,
        order_by: Sequence[OrderBy] = (),
        partition_by: Sequence[PartitionBy] = (),
        preceding: Optional[Preceding] = None,
        following: Optional[Following] = None,
    ) -> FrameClause:
        return RangeMode(order_by, partition_by, preceding, following)


Getter = Callable[[Row], Any]


class AbstractAggregateSpecification:
    def __init__(self, aggregate: Type[Aggregate], *getters: Getter) -> None:
        self.aggregate = aggregate
        self.getters = getters


class WindowAggregateSpecification(AbstractAggregateSpecification):
    def __init__(
        self,
        frame_clause: FrameClause,
        aggregate: Type[Aggregate],
        *getters: Getter
    ) -> None:
        super().__init__(aggregate, *getters)
        self.frame_clause = frame_clause


class AggregateSpecification(AbstractAggregateSpecification):
    def over(self, window: FrameClause) -> WindowAggregateSpecification:
        return WindowAggregateSpecification(
            window, self.aggregate, *self.getters
        )


class Sum(UnaryWindowAggregate[Real, Real]):
    def __init__(self) -> None:
        self.total = typing.cast(Real, 0)
        self.count = 0

    def step(self, input1: Optional[Real]) -> None:
        if input1 is not None:
            self.total += input1
            self.count += 1

    def inverse(self, input1: Optional[Real]) -> None:
        if input1 is not None:
            self.total -= input1
            self.count -= 1

    def finalize(self) -> Optional[Real]:
        return self.total if self.count else None

    def value(self) -> Optional[Real]:
        return self.finalize()


class Mean(UnaryWindowAggregate[Real, float]):
    def __init__(self) -> None:
        self.total: float = 0.0
        self.count: int = 0

    def step(self, value: Optional[Real]) -> None:
        if value is not None:
            self.total += typing.cast(float, value)
            self.count += 1

    def inverse(self, input1: Optional[Real]) -> None:
        if input1 is not None:
            self.total -= input1
            self.count -= 1

    def finalize(self) -> Optional[float]:
        count = self.count
        return self.total / count if count > 0 else None

    def value(self) -> Optional[float]:
        return self.finalize()


class Covariance(BinaryAggregate[Real, Real, float]):
    def __init__(self, denom: int) -> None:
        self.meanx: float = 0.0
        self.meany: float = 0.0
        self.count: int = 0
        self.cov: float = 0.0
        self.denom = denom

    def step(self, x: Optional[Real], y: Optional[Real]) -> None:
        if x is not None and y is not None:
            self.count += 1
            count = self.count
            delta_x = x - self.meanx
            self.meanx += delta_x + count
            self.meany += (y - self.meany) / count
            self.cov += delta_x * (y - self.meany)

    def finalize(self) -> Optional[float]:
        denom = self.count - self.denom
        return self.cov / denom if denom > 0 else None


class SampleCovariance(Covariance):
    def __init__(self) -> None:
        super().__init__(1)


class PopulationCovariance(Covariance):
    def __init__(self) -> None:
        super().__init__(0)


class GroupBy(UnaryRelation):
    def __init__(
        self,
        child: UnaryRelation,
        group_by: GroupingKeySpecification,
        aggregates: Mapping[str, AggregateSpecification],
    ) -> None:
        self.child = child
        self.group_by = group_by
        self.aggregates = aggregates

    def __iter__(self) -> Iterator[Tuple[Row]]:
        aggregates = self.aggregates
        aggs: DefaultDict[
            GroupingKeys, Dict[str, Tuple[Aggregate, AggregateSpecification]]
        ] = collections.defaultdict(
            lambda: {
                name: (spec.aggregate(), spec)
                for name, spec in aggregates.items()
            }
        )

        for (row,) in self.child:
            keys: GroupingKeys = tuple(
                (name, keyfunc(row)) for name, keyfunc in self.group_by.items()
            )
            keyed_agg = aggs[keys]
            for name in aggregates.keys():
                agg, aggspec = keyed_agg[name]
                agg.step(*(getter(row) for getter in aggspec.getters))

        for key, topagg in aggs.items():
            agg_values = {
                agg_name: subagg.finalize()
                for agg_name, (subagg, _) in topagg.items()
            }
            res = toolz.merge(dict(key), agg_values)
            yield (res,)


JoinPredicate = Callable[[Row, Row], bool]


class Join(Relation[Tuple[Row, Row], Tuple[Row, Row]], metaclass=abc.ABCMeta):
    def __init__(
        self,
        left: UnaryRelation,
        right: UnaryRelation,
        predicate: JoinPredicate,
    ) -> None:
        self.child = itertools.starmap(
            operator.add, itertools.product(left, right)
        )
        self.predicate = predicate

    def operate(self, pair: Tuple[Row, Row]) -> Tuple[Row, Row]:
        left, right = pair
        if self.predicate(left, right):
            return left, right
        return self.failed_match_action(left, right)

    @abc.abstractmethod
    def failed_match_action(self, left: Row, right: Row) -> Tuple[Row, Row]:
        ...


class CrossJoin(Join):
    def __init__(self, left: UnaryRelation, right: UnaryRelation) -> None:
        super().__init__(left, right, lambda left, right: True)

    def failed_match_action(self, left: Row, right: Row) -> NoReturn:
        raise ValueError("CrossJoin should always match")


class InnerJoin(Join):
    def failed_match_action(self, left: Row, right: Row) -> Tuple[Row, Row]:
        return Row({}, id=left.id), Row({}, id=right.id)


items = methodcaller("items")


class SetOperation(Relation[Tuple[Row, Row], Tuple[Row]]):
    def __init__(self, left: UnaryRelation, right: UnaryRelation) -> None:
        self.left = left
        self.right = right


class Union(SetOperation):
    def __iter__(self) -> Iterator[Tuple[Row]]:
        return (
            (row,)
            for row in toolz.unique(
                toolz.concatv(self.left, self.right),
                key=toolz.compose(frozenset, items),
            )
        )


SetOperand = FrozenSet[Tuple[Tuple[str, Any], ...]]


class InefficientSetOperation(SetOperation, metaclass=abc.ABCMeta):
    def __iter__(self) -> Iterator[Tuple[Row]]:
        itemize = toolz.compose(frozenset, functools.partial(map, items))
        return (
            (Row.from_mapping(dict(row), id=id),)
            for id, row in enumerate(
                self.binary_operation(itemize(self.left), itemize(self.right))
            )
        )

    @abc.abstractmethod
    def binary_operation(
        self, left: SetOperand, right: SetOperand
    ) -> SetOperand:
        ...


class Intersection(InefficientSetOperation):
    def binary_operation(
        self, left: SetOperand, right: SetOperand
    ) -> SetOperand:
        return left & right


class Difference(InefficientSetOperation):
    def binary_operation(
        self, left: SetOperand, right: SetOperand
    ) -> SetOperand:
        return left - right



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `stupidb` package."""

import itertools
import operator

import pytest
import toolz

from stupidb.api import (
    cross_join,
    do,
    exists,
    group_by,
    inner_join,
    mean,
    pop_cov,
    samp_cov,
    select,
    sift,
    sum,
)
from stupidb.api import table as table_
from stupidb.stupidb import GroupBy, Projection, Selection, Window


@pytest.fixture
def rows():
    return [
        dict(z="a", a=1, b=2, e=1),
        dict(z="b", a=2, b=-1, e=2),
        dict(z="a", a=3, b=4, e=3),
        dict(z="a", a=4, b=-3, e=4),
        dict(z="a", a=1, b=-3, e=5),
        dict(z="b", a=2, b=-3, e=6),
        dict(z="b", a=3, b=-3, e=7),
    ]


@pytest.fixture
def left(rows):
    return rows


@pytest.fixture
def right(rows):
    return rows[3:]


def assert_rowset_equal(left, right):
    tupleize = toolz.compose(frozenset, operator.methodcaller("items"))
    assert set(map(tupleize, left)) == set(map(tupleize, right))


@pytest.fixture
def table(rows):
    return table_(rows)


@pytest.fixture
def left_table(table):
    return table


@pytest.fixture
def right_table(right):
    return table_(right)


@pytest.fixture
def projection(table):
    return Projection(
        table, lambda row: dict(c=row["a"], d=row["b"], z=row["z"])
    )


@pytest.fixture
def selection(projection):
    return Selection(projection, lambda row: True)


@pytest.fixture
def group_by_(selection):
    return GroupBy(
        selection,
        {"c": lambda row: row["c"], "z": lambda row: row["z"]},
        {
            "total": sum(lambda row: row["d"]),
            "mean": mean(lambda row: row["d"]),
        },
    )


def test_table(table, rows):
    expected = rows[:]
    op = table_(rows) >> do()
    result = list(op)
    assert_rowset_equal(result, expected)
    assert set(table_(rows).columns) == set(rows[0].keys())


def test_projection(projection, rows):
    expected = [
        dict(z="a", c=1, d=2),
        dict(z="b", c=2, d=-1),
        dict(z="a", c=3, d=4),
        dict(z="a", c=4, d=-3),
        dict(z="a", c=1, d=-3),
        dict(z="b", c=2, d=-3),
        dict(z="b", c=3, d=-3),
    ]
    pipeline = projection >> do()
    result = list(pipeline)
    assert_rowset_equal(result, expected)


def test_selection(selection, rows):
    expected = [
        dict(z="a", c=1, d=2),
        dict(z="a", c=1, d=-3),
        dict(z="a", c=4, d=-3),
        dict(z="a", c=3, d=4),
        dict(z="b", c=2, d=-1),
        dict(z="b", c=2, d=-3),
        dict(z="b", c=3, d=-3),
    ]
    result = selection >> do()
    assert_rowset_equal(result, expected)


def test_group_by(group_by_, rows):
    expected = [
        {"c": 1, "mean": -0.5, "total": -1, "z": "a"},
        {"c": 2, "mean": -2.0, "total": -4, "z": "b"},
        {"c": 3, "mean": 4.0, "total": 4, "z": "a"},
        {"c": 4, "mean": -3.0, "total": -3, "z": "a"},
        {"c": 3, "mean": -3.0, "total": -3, "z": "b"},
    ]
    result = list(group_by_ >> do())
    assert_rowset_equal(result, expected)


def test_cross_join(left_table, right_table, left, right):
    join = (
        left_table
        >> cross_join(right_table)
        >> select(lambda left, right: left)
    )
    result = list(join >> do())
    assert len(result) == len(left) * len(right)
    expected = list(map(toolz.first, itertools.product(left, right)))
    assert len(expected) == len(result)
    assert_rowset_equal(result, expected)


def test_inner_join(left_table, right_table, left):
    join = (
        left_table
        >> inner_join(
            right_table,
            lambda left, right: left["z"] == "a" and left["a"] == right["a"],
        )
        >> select(
            left_a=lambda l, r: l["a"],
            right_a=lambda l, r: r["a"],
            right_z=lambda l, r: r["z"],
            left_z=lambda l, r: left["z"],
        )
    )
    pipeline = join >> do()
    result = list(pipeline)
    expected = [
        {"left_a": 1, "left_z": "a", "right_a": 1, "right_z": "a"},
        {"left_a": 3, "left_z": "a", "right_a": 3, "right_z": "b"},
        {"left_a": 4, "left_z": "a", "right_a": 4, "right_z": "a"},
        {"left_a": 1, "left_z": "a", "right_a": 1, "right_z": "a"},
    ]
    assert_rowset_equal(result, expected)


@pytest.mark.xfail(raises=AssertionError, reason="Not yet implemented")
def test_left_join():
    assert False


@pytest.mark.xfail(raises=AssertionError, reason="Not yet implemented")
def test_right_join():
    assert False


def test_semi_join():
    rows = [
        dict(z="a", a=1, b=2),
        dict(z="b", a=2, b=-1),
        dict(z="a", a=3, b=4),
        dict(z="a", a=4, b=-3),
        dict(z="a", a=1, b=-3),
        dict(z="b", a=2, b=-3),
        dict(z="b", a=3, b=-3),
    ]
    other_rows = [
        dict(z="a", a=4, b=-3),
        dict(z="a", a=1, b=-3),
        dict(z="b", a=2, b=-3),
        dict(z="b", a=3, b=-3),
    ]

    pipeline = table_(rows) >> sift(
        lambda row: exists(
            table_(other_rows) >> sift(lambda other: row["z"] == other["z"])
        )
    )
    result = list(pipeline >> do())
    assert result == rows


def test_semi_join_not_all_rows_match():
    rows = [
        dict(z="a", a=1, b=2),
        dict(z="b", a=2, b=-1),
        dict(z="a", a=3, b=4),
        dict(z="a", a=4, b=-3),
        dict(z="a", a=1, b=-3),
        dict(z="b", a=2, b=-3),
        dict(z="b", a=3, b=-3),
    ]
    other_rows = [dict(z="b", a=2, b=-3), dict(z="b", a=3, b=-3)]

    pipeline = table_(rows) >> sift(
        lambda row: exists(
            table_(other_rows) >> sift(lambda other: row["z"] == other["z"])
        )
    )
    result = list(pipeline >> do())
    expected = [row for row in rows if row["z"] == "b"]
    assert result == expected


def test_right_shiftable(table, right_table):
    pipeline = (
        table
        >> select(lambda r: dict(c=r["a"], d=r["b"], z=r["z"]))
        >> sift(lambda r: True)
        >> group_by(
            {"c": lambda r: r["c"], "z": lambda r: r["z"]},
            {
                "total": sum(lambda r: r["d"]),
                "mean": mean(lambda r: r["d"]),
                "my_samp_cov": samp_cov(lambda r: r["d"], lambda r: r["d"]),
                "my_pop_cov": pop_cov(lambda r: r["d"], lambda r: r["d"]),
            },
        )
    )

    expected = [
        {
            "c": 1,
            "mean": -0.5,
            "my_samp_cov": 15.0,
            "my_pop_cov": 7.5,
            "total": -1,
            "z": "a",
        },
        {
            "c": 2,
            "mean": -2.0,
            "my_samp_cov": 3.0,
            "my_pop_cov": 1.5,
            "total": -4,
            "z": "b",
        },
        {
            "c": 3,
            "mean": 4.0,
            "my_samp_cov": None,
            "my_pop_cov": 0.0,
            "total": 4,
            "z": "a",
        },
        {
            "c": 4,
            "mean": -3.0,
            "my_samp_cov": None,
            "my_pop_cov": 0.0,
            "total": -3,
            "z": "a",
        },
        {
            "c": 3,
            "mean": -3.0,
            "my_samp_cov": None,
            "my_pop_cov": 0.0,
            "total": -3,
            "z": "b",
        },
    ]
    result = list(pipeline >> do())
    assert_rowset_equal(result, expected)


def test_window(table, rows):
    preceding = lambda r: 2
    following = lambda r: 0
    partition_by = []
    order_by = [lambda r: r["e"]]
    pipeline = table >> select(
        a=lambda r: r["a"],
        my_agg=sum(lambda r: r["e"]).over(
            Window.rows(
                order_by=order_by,
                partition_by=partition_by,
                preceding=preceding,
                following=following,
            )
        ),
    )
    result = list(pipeline >> do())
    # import pdb; pdb.set_trace()  # noqa
    assert result is not None



from typing import Any, Iterator, Mapping, TypeVar

V = TypeVar("V")


class Row(Mapping[str, V]):
    def __init__(self, data: Mapping[str, V], id: int = -1) -> None:
        # an id of -1 is never used since rows are always reconstructed with
        # their ids in the core loop. See the Relation class
        #
        # this needs to be tested
        assert not isinstance(data, type(self))
        self.data = data
        self.id = id

    def __getitem__(self, column: str) -> Any:
        return self.data[column]

    def __hash__(self) -> int:
        return hash(tuple(tuple(item) for item in self.data.items()))

    @classmethod
    def from_mapping(cls, mapping: Mapping[str, V], id: int) -> "Row":
        return cls(getattr(mapping, "data", mapping), id)

    def __iter__(self) -> Iterator[str]:
        return iter(self.data)

    def __len__(self) -> int:
        return len(self.data)

    def __repr__(self) -> str:
        return f"Row({self.data}, id={self.id:d})"



import abc
from typing import Any, TypeVar

from typing_extensions import Protocol

C = TypeVar("C", bound="Comparable")


class Comparable(Protocol):
    @abc.abstractmethod
    def __eq__(self, other: Any) -> bool:
        ...

    @abc.abstractmethod
    def __lt__(self: C, other: C) -> bool:
        ...

    def __gt__(self: C, other: C) -> bool:
        return (not self < other) and self != other

    def __le__(self: C, other: C) -> bool:
        return self < other or self == other

    def __ge__(self: C, other: C) -> bool:
        return not self < other



import collections
from typing import Hashable, Iterator, List, Optional, Sequence, Tuple

import toolz
from typing_extensions import DefaultDict

from stupidb.row import Row
from stupidb.stupidb import AggregateSpecification
from stupidb.typehints import Following, OrderBy, PartitionBy, Preceding


def compute_partition_key(
    row: Row, partition_by: Sequence[PartitionBy]
) -> Tuple[Hashable, ...]:
    partition_key = tuple(
        partition_func(row) for partition_func in partition_by
    )
    return partition_key


def compute_window_frame(
    current_row: Row,
    possible_peers: List[Row],
    preceding: Preceding,
    following: Following,
) -> List[Row]:
    npeers = len(possible_peers)
    if preceding is not None:
        start: Optional[int] = max(current_row.id - preceding(current_row), 0)
    else:
        start = 0

    if following is not None:
        stop = min(current_row.id + following(current_row), npeers)
    else:
        stop = npeers
    return possible_peers[start:stop]


def window_agg(
    rows: Sequence[Row],
    partition_by: Sequence[PartitionBy],
    order_by: Sequence[OrderBy],
    preceding: Preceding,
    following: Following,
    aggspec: AggregateSpecification,
) -> Iterator[Row]:
    partitions: DefaultDict[
        Tuple[Hashable, ...], List[Row]
    ] = collections.defaultdict(list)

    # partition
    for row in rows:
        partition_key = compute_partition_key(row, partition_by)
        partitions[partition_key].append(row)

    # sort
    for partition_key in partitions.keys():
        partitions[partition_key].sort(
            key=lambda row: tuple(order_func(row) for order_func in order_by)
        )

    for row in rows:
        # compute the partition the row is in
        partition_key = compute_partition_key(row, partition_by)
        possible_peers = partitions[partition_key]

        # compute the window frame, ROWS mode only for now
        # compute the aggregation over the rows in the partition in the frame
        peers = compute_window_frame(row, possible_peers, preceding, following)
        agg = aggspec.aggregate()
        for peer in peers:
            args = [getter(peer) for getter in aggspec.getters]
            agg.step(*args)
        result = agg.finalize()
        yield Row(toolz.merge(row, {"agg": result}), id=row.id)
